/* 

	pmmixer.c

	Sound Blaster Pro Mixer for PM

	by David Nichols, written for GCC/2

	for use with the SBOS2.SYS driver by Michael Fulbright

*/

#include <stdio.h>

#define INCL_DOS
#define INCL_DOSDEVIOCTL /* for emx */
#define INCL_WIN
#define INCL_GPI

#include <stdlib.h>
#include <os2.h>
#include <string.h>
#include <errno.h>
#include "pmmixer.h"
#include "sblast_user.h"

#define NUMSCROLLBARS 11

typedef struct _REAL_SWCNTRL          /* swctl from C Set/2 */
{
   HWND     hwnd;
   HWND     hwndIcon;
   HPROGRAM hprog;
   PID      idProcess;
   ULONG    idSession;
   ULONG    uchVisibility;
   ULONG    fbJump;
   CHAR     szSwtitle[MAXNAMEL+4];
   ULONG    bProgType;
} REAL_SWCNTRL;


struct volumescrollbarrec
{
   ULONG id;
   short current;
   short max;
};

struct volumescrollbarrec scrollbars[NUMSCROLLBARS] = 
{ ID_MASTERLEFT, 0, 14,	ID_MASTERRIGHT, 0, 14,
  ID_VOCLEFT, 0, 14,	ID_VOCRIGHT, 0, 14,
  ID_FMLEFT, 0, 14,	ID_FMRIGHT, 0, 14,
  ID_LINELEFT, 0, 14,	ID_LINERIGHT, 0, 14,
  ID_CDLEFT, 0, 14,     ID_CDRIGHT, 0 ,14,
  ID_MIC, 0, 6 };

struct sb_mixer_levels sblevels;
struct sb_mixer_params sbparams;

HFILE mixer_handle;

HWND hwndMixer = NULL;

char title[] = "SB Pro Mixer";

int mute = 0;
int locklr = 0;
int filterhigh = 0;
int filterin = 0;
int filterout = 0;
int recsrc = SRC_LINE;

int InitSB()
{
   ULONG action;

   return !DosOpen("SBMIX$", &mixer_handle, &action, 0, FILE_NORMAL, FILE_OPEN, 
	OPEN_ACCESS_READWRITE|OPEN_SHARE_DENYNONE|OPEN_FLAGS_WRITE_THROUGH,
	NULL);
}

void ReadSBLevels()
{
   ULONG parlen = 0;
   ULONG datlen = sizeof(struct sb_mixer_levels);

   DosDevIOCtl(mixer_handle, DSP_CAT, MIXER_IOCTL_READ_LEVELS,
		NULL, 0, &parlen, &sblevels, datlen,
		&datlen);

   if (!mute)
   {
      scrollbars[0].current = sblevels.master.l;
      scrollbars[1].current = sblevels.master.r;
   }
   scrollbars[2].current = sblevels.voc.l;
   scrollbars[3].current = sblevels.voc.r;
   scrollbars[4].current = sblevels.fm.l;
   scrollbars[5].current = sblevels.fm.r;
   scrollbars[6].current = sblevels.line.l;
   scrollbars[7].current = sblevels.line.r;
   scrollbars[8].current = sblevels.cd.l;
   scrollbars[9].current = sblevels.cd.r;
   scrollbars[10].current = sblevels.mic;
}

void SetSBLevels()
{
   ULONG parlen, datlen;

   datlen = sizeof(struct sb_mixer_levels);
   parlen = 0;
   if (mute)
   {
      sblevels.master.l = 0;
      sblevels.master.r = 0;
   }
   else
   {
      sblevels.master.l = scrollbars[0].current;
      sblevels.master.r = scrollbars[1].current;
   }
   sblevels.voc.l = scrollbars[2].current;
   sblevels.voc.r = scrollbars[3].current;
   sblevels.fm.l = scrollbars[4].current;
   sblevels.fm.r = scrollbars[5].current;
   sblevels.line.l = scrollbars[6].current;
   sblevels.line.r = scrollbars[7].current;
   sblevels.cd.l = scrollbars[8].current;
   sblevels.cd.r = scrollbars[9].current;
   sblevels.mic = scrollbars[10].current;
   DosDevIOCtl(mixer_handle, DSP_CAT, MIXER_IOCTL_SET_LEVELS,
               NULL, 0, &parlen, &sblevels, datlen, &datlen);
}

void SetSBParams()
{
   ULONG datlen, parlen;

   parlen = 0;
   datlen = sizeof(struct sb_mixer_params);                     

   sbparams.hifreq_filter = filterhigh;
   sbparams.filter_output = filterout;
   sbparams.filter_input  = filterin;  
   sbparams.record_source = recsrc;
   DosDevIOCtl(mixer_handle, DSP_CAT, MIXER_IOCTL_SET_PARAMS,
              NULL, 0, &parlen,                              
              &sbparams, datlen, &datlen);                   
}

void ReadSBParams()
{
   ULONG datlen, parlen;

   parlen = 0;
   datlen = sizeof(struct sb_mixer_params);                     
   DosDevIOCtl(mixer_handle, DSP_CAT, MIXER_IOCTL_READ_PARAMS,
              NULL, 0, &parlen,                              
              &sbparams, datlen, &datlen);
   filterhigh = sbparams.hifreq_filter;
   filterout  = sbparams.filter_output;
   filterin   = sbparams.filter_input;                   
   recsrc     = sbparams.record_source;
}

void UpdateButtons(HWND hwnd)
{
   WinSendDlgItemMsg(hwnd, ID_FILTERHIGH, BM_SETCHECK, (MPARAM)filterhigh, (MPARAM)0);
   WinSendDlgItemMsg(hwnd, ID_FILTERLOW, BM_SETCHECK, (MPARAM)!filterhigh, (MPARAM)0);
   WinSendDlgItemMsg(hwnd, ID_FILTERIN, BM_SETCHECK, (MPARAM)filterin, (MPARAM)0);
   WinSendDlgItemMsg(hwnd, ID_FILTEROUT, BM_SETCHECK, (MPARAM)filterout, (MPARAM)0);
   WinSendDlgItemMsg(hwnd, ID_RECORDLINE, BM_SETCHECK, (MPARAM)(recsrc==SRC_LINE), (MPARAM)0);
   WinSendDlgItemMsg(hwnd, ID_RECORDCD, BM_SETCHECK, (MPARAM)(recsrc==SRC_CD), (MPARAM)0);
   WinSendDlgItemMsg(hwnd, ID_RECORDMIC, BM_SETCHECK, (MPARAM)(recsrc==SRC_MIC), (MPARAM)0);
}

void UpdateScrollPos(HWND hwnd)
{
   int i;

   for(i=0;i<NUMSCROLLBARS;i++)
      WinSendDlgItemMsg(hwnd, scrollbars[i].id, SBM_SETSCROLLBAR,
	MPFROM2SHORT(scrollbars[i].max+1-scrollbars[i].current,0),
	MPFROM2SHORT(0,scrollbars[i].max));
}

void StoreNewMixerLevel(USHORT id, USHORT newpos)
{
   ULONG i = 0;

   while (((int)i<NUMSCROLLBARS)&&(id!=scrollbars[i].id)) i++;
   if (i<NUMSCROLLBARS)
   {
      scrollbars[i].current = scrollbars[i].max+1-newpos;
      if (locklr&&(i<NUMSCROLLBARS-1))
      {
	 if (!(i%2)) i++;
	 else i--;
	 scrollbars[i].current = scrollbars[i].max+1-newpos;
      } 
   }
}

void LockVolume(HWND hwnd)
{
   int i;

   for(i=0;i<NUMSCROLLBARS-1;i+=2)
      scrollbars[i].current = scrollbars[i+1].current;
   SetSBLevels();
   UpdateScrollPos(hwnd);
}

#pragma linkage(MixerDlgProc, system)

MRESULT MixerDlgProc(HWND hwnd, USHORT msg, MPARAM mp1, MPARAM mp2)
{
   static USHORT id, mess;
   static ULONG slidepos;
   static MRESULT mr;
   static USHORT smin, smax;
   int update = 0;

   switch(msg)
   {
      case WM_INITDLG:
         if (!InitSB()) 
         {
            WinMessageBox(HWND_DESKTOP, hwnd, "Unable to initialize Sound Blaster device.",title, 0, MB_ICONEXCLAMATION|MB_MOVEABLE|MB_OK);
      	    exit(1);
   	 }
	 ReadSBParams();
	 ReadSBLevels();
	 UpdateButtons(hwnd);
	 UpdateScrollPos(hwnd);
	 WinSendDlgItemMsg(hwnd, ID_LOCKVOL, BM_SETCHECK, (MPARAM)1, (MPARAM)0);
	 WinSendMsg(hwnd, WM_CONTROL, MPFROM2SHORT(ID_LOCKVOL, BN_CLICKED), (MPARAM)0);
 	 break;

      case WM_ACTIVATE:
	 ReadSBParams();
	 ReadSBLevels();
	 UpdateButtons(hwnd);
	 UpdateScrollPos(hwnd);
	 break;
	
      case WM_CONTROL:
	 id = SHORT1FROMMP(mp1);
	 mess = SHORT2FROMMP(mp1);
	 if ((mess == BN_CLICKED)||(mess == BN_DBLCLICKED))
 	 {
	    switch (id)
            {
	       case ID_LOCKVOL:
 		  locklr = !locklr;
	       	  if (locklr) LockVolume(hwnd);
   	          break;

	       case ID_MUTE:
		  mute = !mute;
		  SetSBLevels(); 
		  UpdateScrollPos(hwnd);
	 	  break;

	       case ID_FILTERIN:
		  filterin = !filterin;
		  SetSBParams();
		  break;

	       case ID_FILTEROUT:
		  filterout = !filterout;
		  SetSBParams();
		  break;

	       case ID_FILTERHIGH:
		  if (!filterhigh)
		     WinSendDlgItemMsg(hwnd, ID_FILTERLOW, BM_SETCHECK, 0, 0);
		  WinSendDlgItemMsg(hwnd, ID_FILTERHIGH, BM_SETCHECK, (MPARAM)1, 0);
	 	  filterhigh = TRUE;
		  SetSBParams();
		  break;

	       case ID_FILTERLOW:
		  if (filterhigh)
		     WinSendDlgItemMsg(hwnd, ID_FILTERHIGH, BM_SETCHECK, 0, 0);
		  WinSendDlgItemMsg(hwnd, ID_FILTERLOW, BM_SETCHECK, (MPARAM)1, 0);
		  filterhigh = FALSE;
		  SetSBParams();
		  break;

	       case ID_RECORDLINE:
		  recsrc = SRC_LINE;
		  SetSBParams();
		  break;

	       case ID_RECORDCD:
		  recsrc = SRC_CD;
		  SetSBParams();
		  break;

	       case ID_RECORDMIC:
		  recsrc = SRC_MIC;
		  SetSBParams();
		  break;
	    }
	 }
         break;

      case WM_VSCROLL:
	 id = SHORT1FROMMP(mp1);
	 if (id)
	 {
	    ReadSBLevels();
	    UpdateScrollPos(hwnd);
	    update = 1;
	    switch (SHORT2FROMMP(mp2))
	    {
               case SB_LINEUP:

               case SB_LINEDOWN:
                  slidepos = (ULONG)WinSendDlgItemMsg(hwnd, id, SBM_QUERYPOS,
                                        (MPARAM)0, (MPARAM)0);
                  mr = WinSendDlgItemMsg(hwnd, id, SBM_QUERYRANGE, (MPARAM)0, (MPARAM)0);
                  if (SHORT2FROMMP(mp2)==SB_LINEUP)
                  {
                     smin=SHORT1FROMMR(mr);
                     slidepos-=2;
                     if (slidepos < smin) slidepos=smin;
                  }
                  else
                  {
                     smax=SHORT2FROMMR(mr);
                     slidepos+=2;
                     if (slidepos > smax) slidepos=smax;
                  }
                  break;

               case SB_PAGEUP:
               
	       case SB_PAGEDOWN:
                  slidepos = (ULONG)WinSendDlgItemMsg(hwnd, id, SBM_QUERYPOS,
                                          (MPARAM)0, (MPARAM)0);

                  mr = WinSendDlgItemMsg(hwnd, id, SBM_QUERYRANGE, (MPARAM)0, (MPARAM)0);
                  if (SHORT2FROMMP(mp2)==SB_PAGEUP)
                  {
                     smin=SHORT1FROMMR(mr);
                     slidepos -=4;
                     if (slidepos < smin) slidepos=smin;
                  }
                  else
                  {
                     smax=SHORT2FROMMR(mr);
                     slidepos +=4;
                     if (slidepos >smax) slidepos=smax;
                  }
                  break;

               case SB_SLIDERPOSITION:
            
	       case SB_SLIDERTRACK:
		  slidepos=SHORT1FROMMP(mp2);
                  break;

               default:
                  update=0;
                  break;
            }

            if (update)  
            {
               /* send new volume to SBMIX */
               StoreNewMixerLevel(id, slidepos);
	       SetSBLevels();
	       UpdateScrollPos(hwnd);
	    }
	 }
	 break;

      case WM_CLOSE:
 	   return WinDefWindowProc(hwnd, msg, mp1, mp2);

      case WM_COMMAND:
 	 switch(SHORT1FROMMP(mp1))
  	 {
	    case IDM_REFRESH:
	       ReadSBParams();
	       ReadSBLevels();
	       UpdateButtons(hwnd);
	       UpdateScrollPos(hwnd);
	       return 0;

            default:
	       return (MRESULT)0;
	 }
	 break;
   }
   return WinDefDlgProc(hwnd, msg, mp1, mp2);
}

void main()
{
   char szClientClass[] = "mixer.child";
   HAB hab;
   HMQ hmq;
   QMSG qmsg;
   HPOINTER hptr;
   HSWITCH hSwitch;                  /* Switch entry handle        */
   REAL_SWCNTRL SwitchData;               /* Switch control data block  */
   PID pid;

   hab = WinInitialize (0) ;
   hmq = WinCreateMsgQueue (hab, 0) ;

   WinRegisterClass (hab, szClientClass, MixerDlgProc, 0, 0);

   hwndMixer = WinLoadDlg(HWND_DESKTOP, NULL, MixerDlgProc, NULL, IDM_MIXERDLG, NULL);

   WinQueryWindowProcess(hwndMixer, &pid, NULL);

   SwitchData.hwnd = hwndMixer;               /* Set frame window handle    */
   SwitchData.hwndIcon = NULL;                   /* Use default icon           */
   SwitchData.hprog = NULL;                      /* Use default program handle */
   SwitchData.idProcess = pid;                  /* Use current process id     */
   SwitchData.idSession = 0;                  /* Use current session id     */
   SwitchData.uchVisibility = SWL_VISIBLE;    /* Make visible               */
   SwitchData.fbJump = SWL_JUMPABLE;          /* Make jumpable via Alt+Esc  */
   SwitchData.bProgType = PROG_PM;            /* program type */
   strcpy(SwitchData.szSwtitle,title);        /* title text                 */

   hSwitch = WinCreateSwitchEntry(hab, (PSWCNTRL)&SwitchData);  /* Add switch entry   */

   hptr = WinLoadPointer(HWND_DESKTOP, NULL, ID_MIXER);
   WinPostMsg(hwndMixer, WM_SETICON, hptr, NULL);
 
   while (WinGetMsg (hab, &qmsg, NULL, 0, 0))
       WinDispatchMsg (hab, &qmsg);

   WinRemoveSwitchEntry(hSwitch);
   WinDestroyWindow(hwndMixer); 
   WinDestroyMsgQueue(hmq);
   WinTerminate(hab);
}
